@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 #
-# Copyright 2018-2021 Michael Samoglyadov
+# Copyright 2018-2022 Michael Samoglyadov
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -14,18 +14,15 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-"""
-Recursive diff for nested structures, implementation of
-https://github.com/mr-mixas/Nested-Diff
+"""Recursive diff and patch for nested structures."""
 
-"""
 from difflib import SequenceMatcher
 from pickle import dumps
-
+from warnings import warn
 
 __all__ = ['Differ', 'Iterator', 'Patcher', 'diff', 'patch']
 
-__version__ = '0.7'
+__version__ = '0.9'
 __author__ = 'Michael Samoglyadov'
 __license__ = 'Apache License, Version 2.0'
 __website__ = 'https://github.com/mr-mixas/Nested-Diff.py'
@@ -48,13 +45,12 @@ class Differ(object):
     Diff metadata alternates with actual data; simple types specified as is,
     dicts, lists and tuples contain subdiffs for their items with native for
     such types addressing: indexes for lists and tuples and keys for
-    dictionaries. Each status type, except `D`. `E` and `I`, may be omitted
+    dictionaries. Any status type, except `D`, `E` and `I` may be omitted
     during diff computation. `E` tag is used with `D` when entity unable to
     contain diff by itself (set, frozenset); `D` contain a list of subdiffs
     in this case.
 
     Example:
-
     a:  {"one": [5,7]}
     b:  {"one": [5], "two": 2}
     opts: U=False  # omit unchanged items
@@ -83,8 +79,10 @@ class Differ(object):
     compared by values.
 
     """
-    def __init__(self, A=True, N=True, O=True, R=True, U=True,  # noqa: E741
-                 trimR=False, diff_method=None, multiline_diff_context=-1):
+
+    def __init__(self, A=True, N=True, O=True, R=True, U=True,  # noqa: E501 E741 N803
+                 trimR=False, diff_method=None, text_diff_ctx=-1,
+                 multiline_diff_context=-1):
         """
         Construct Differ.
 
@@ -96,12 +94,13 @@ class Differ(object):
         default is `False`.
 
         `diff_method` method with such name (if object have one) from first
-        diffed object will be called for diff. Second diffed object and all
-        Differ opts will be passed as arguments, diff expected for output.
-        Disabled (`None`) by default.
+        diffed object will be called for diff. Second diffed object and current
+        differ will be passed as arguments, diff expected for output.
+        Disabled (`None`) by default. This option is deprecated and will be
+        removed in the next release.
 
-        `multiline_diff_context` defines amount of context lines for multiline
-        string diffs, multiline diffs disabled when value is negative.
+        `text_diff_ctx` defines amount of context lines for text (multiline
+        strings) diffs, disabled entirely when value is negative.
 
         """
         self.__diff_method = diff_method
@@ -123,38 +122,38 @@ class Differ(object):
         }
 
         if multiline_diff_context >= 0:
-            self.__differs[str] = self.diff_multiline
-            self.multiline_diff_context = multiline_diff_context
+            warn('`multiline_diff_context` is deprecated and will be removed'
+                 ' in the next release, use `text_diff_ctx` instead',
+                 DeprecationWarning, stacklevel=2)
+            if text_diff_ctx < 0:
+                text_diff_ctx = multiline_diff_context
+            else:
+                raise TypeError('`multiline_diff_context` and `text_diff_ctx`'
+                                ' can not be used together')
+
+        if text_diff_ctx >= 0 and self.op_n and self.op_o:
+            self.__differs[str] = self.diff_text
+            self.text_diff_ctx = text_diff_ctx
 
     def diff(self, a, b):
         """
-        Compute diff for two arbitrary objects.
+        Return diff for two arbitrary objects.
 
         This method is a dispatcher and calls registered diff method for each
         diffed values pair according to their type. `diff__default` called for
-        unequal and not registered types. Args and kwargs passed to called
-        method as is.
+        unequal and not registered types. Args passed to called method as is.
 
         :param a: First object to diff.
         :param b: Second object to diff.
 
         """
         if self.__diff_method is not None:
+            warn('diff internal methods are deprecated and will be removed in'
+                 ' the next release', DeprecationWarning, stacklevel=2)
             try:
-                method = a.__getattribute__(self.__diff_method)
+                return a.__getattribute__(self.__diff_method)(b, self)
             except AttributeError:
                 pass
-            else:
-                return method(
-                    b,
-                    A=self.op_a,
-                    N=self.op_n,
-                    O=self.op_o,  # noqa: E741
-                    R=self.op_r,
-                    U=self.op_u,
-                    trimR=self.op_trim_r,
-                    diff_method=self.__diff_method,
-                )
 
         if a.__class__ is b.__class__:
             # it's faster to compare pickled dumps and dig differences
@@ -167,10 +166,7 @@ class Differ(object):
         return self.diff__default(a, b)
 
     def diff__default(self, a, b):
-        """
-        Return default diff.
-
-        """
+        """Return default diff."""
         dif = {}
 
         if self.op_n:
@@ -182,7 +178,7 @@ class Differ(object):
 
     def diff_dict(self, a, b):
         """
-        Compute diff for two dicts.
+        Return diff for two dicts.
 
         :param a: First dict to diff.
         :param b: Second dict to diff.
@@ -190,8 +186,8 @@ class Differ(object):
         >>> a = {'one': 1, 'two': 2, 'three': 3}
         >>> b = {'one': 1, 'two': 42}
         >>>
-        >>> Differ(O=False, U=False).diff_dicts(a, b)
-        {'D': {'two': {'N': 42}, 'three': {'R': 3}}}
+        >>> Differ(O=False, U=False).diff_dict(a, b)
+        {'D': {'three': {'R': 3}, 'two': {'N': 42}}}
         >>>
 
         """
@@ -222,7 +218,7 @@ class Differ(object):
 
     def diff_list(self, a, b):
         """
-        Compute diff for two lists.
+        Return diff for two lists.
 
         :param a: First list to diff.
         :param b: Second list to diff.
@@ -230,7 +226,7 @@ class Differ(object):
         >>> a = [0,1,2,3]
         >>> b = [  1,2,4,5]
         >>>
-        >>> Differ(O=False, U=False).diff_lists(a, b)
+        >>> Differ(O=False, U=False).diff_list(a, b)
         {'D': [{'R': 0}, {'N': 4, 'I': 3}, {'A': 5}]}
         >>>
 
@@ -284,21 +280,25 @@ class Differ(object):
         return {}
 
     def diff_multiline(self, a, b):
-        """
-        Compute diff for multiline strings.
+        warn('Differ.diff_multiline method is deprecated and will be removed'
+             'in the next release', DeprecationWarning, stacklevel=2)
+        return self.diff_text(self, a, b)
 
-        Result is a unified diff formatted as usual nested diff structure with
-        'I' tagged subdiffs to contain hunks headers.
+    def diff_text(self, a, b):
+        r"""
+        Return diff for texts (multiline strings).
+
+        Result is a unified-like diff formatted as nested diff structure, with
+        'I' tagged subdiffs containing hunks headers.
 
         :param a: First string to diff.
         :param b: Second string to diff.
 
-        >>> a = 'A\nB\nC'
-        >>> b = 'A\nC'
+        >>> a = 'one'
+        >>> b = 'one\ntwo'
         >>>
-        >>> Differ(multiline_diff_context=3).diff_multiline(a, b)
-        {'D': [{'I': [0, 3, 0, 2]}, {'U': 'A'}, {'R': 'B'}, {'U': 'C'}],
-         'E': ''}
+        >>> Differ(text_diff_ctx=3).diff_text(a, b)
+        {'D': [{'I': [0, 1, 0, 2]}, {'U': 'one'}, {'A': 'two'}], 'E': ''}
 
         """
         lines_a = a.split('\n', -1)
@@ -311,7 +311,7 @@ class Differ(object):
         self.lcs.set_seq1(lines_a)
         self.lcs.set_seq2(lines_b)
 
-        for group in self.lcs.get_grouped_opcodes(self.multiline_diff_context):
+        for group in self.lcs.get_grouped_opcodes(self.text_diff_ctx):
             dif.append({
                 'I': [
                     group[0][1], group[-1][2],
@@ -321,17 +321,14 @@ class Differ(object):
 
             for op, i1, i2, j1, j2 in group:
                 if op == 'equal':
-                    for line in lines_a[i1:i2]:
-                        dif.append({'U': line})
+                    dif.extend({'U': line} for line in lines_a[i1:i2])
                     continue
 
-                if op in {'replace', 'delete'}:
-                    for line in lines_a[i1:i2]:
-                        dif.append({'R': line})
+                if op != 'insert':
+                    dif.extend({'R': line} for line in lines_a[i1:i2])
 
-                if op in {'replace', 'insert'}:
-                    for line in lines_b[j1:j2]:
-                        dif.append({'A': line})
+                if op != 'delete':
+                    dif.extend({'A': line} for line in lines_b[j1:j2])
 
         if dif:
             return {'D': dif, 'E': a.__class__()}
@@ -340,7 +337,7 @@ class Differ(object):
 
     def diff_set(self, a, b):
         """
-        Compute diff for two [frozen]sets.
+        Return diff for two [frozen]sets.
 
         :param a: First set to diff.
         :param b: Second set to diff.
@@ -348,7 +345,7 @@ class Differ(object):
         >>> a = {1, 2}
         >>> b = {2, 3}
         >>>
-        >>> Differ(U=False).diff_sets(a, b)
+        >>> Differ(U=False).diff_set(a, b)
         {'D': [{'R': 1}, {'A': 3}], 'E': set()}
         >>>
 
@@ -356,15 +353,13 @@ class Differ(object):
         dif = []
 
         for i in a.union(b):
-            if i in a and i in b:
-                if self.op_u:
-                    dif.append({'U': i})
-
-            elif i in a:  # removed
-                if self.op_r:
+            if i in a:
+                if i in b:
+                    if self.op_u:
+                        dif.append({'U': i})
+                elif self.op_r:
                     # ignore trimR opt here: value required for removal
                     dif.append({'R': i})
-
             else:  # added
                 if self.op_a:
                     dif.append({'A': i})
@@ -376,7 +371,7 @@ class Differ(object):
 
     def diff_tuple(self, a, b):
         """
-        Compute diff for two tuples.
+        Return diff for two tuples.
 
         :param a: First tuple to diff.
         :param b: Second tuple to diff.
@@ -384,15 +379,17 @@ class Differ(object):
         >>> a = (  1,2,4,5)
         >>> b = (0,1,2,3)
         >>>
-        >>> Differ(O=False, U=False).diff_tuples(a, b)
+        >>> Differ(O=False, U=False).diff_tuple(a, b)
         {'D': ({'A': 0}, {'N': 3, 'I': 2}, {'R': 5})}
         >>>
 
         """
         dif = self.diff_list(a, b)
 
-        if 'D' in dif:
+        try:
             dif['D'] = tuple(dif['D'])
+        except KeyError:
+            pass
 
         return dif
 
@@ -420,18 +417,18 @@ class Differ(object):
 
 
 class Patcher(object):
-    """
-    Patch objects using nested diff.
+    """Patch objects using nested diff."""
 
-    """
     def __init__(self, patch_method=None):
         """
         Construct Patcher.
 
         Optional arguments:
         `patch_method` method with such name, if patched object have one, will
-        be called with patch as argument. Patched object expected for output.
-        Disabled (`None`) by default.
+        be called with patch and current patcher as arguments. Patched object
+        expected for output.
+        Disabled (`None`) by default. This option is deprecated and will be
+        removed in the next release.
 
         """
         self.__patch_method = patch_method
@@ -469,21 +466,23 @@ class Patcher(object):
 
         """
         if self.__patch_method is not None:
+            warn('patch internal methods are deprecated and will be removed in'
+                 ' the next release', DeprecationWarning, stacklevel=2)
             try:
-                method = target.__getattribute__(self.__patch_method)
+                return target.__getattribute__(self.__patch_method)(
+                    ndiff, self)
             except AttributeError:
                 pass
-            else:
-                return method(ndiff)
 
         if 'D' in ndiff:
             return self.get_patcher(
                 ndiff['E' if 'E' in ndiff else 'D'].__class__,
             )(target, ndiff)
-        elif 'N' in ndiff:
-            return ndiff['N']
 
-        return target
+        try:
+            return ndiff['N']
+        except KeyError:
+            return target
 
     def patch_dict(self, target, ndiff):
         """
@@ -551,16 +550,16 @@ class Patcher(object):
                 idx = subdiff['I'][0] + offset
             elif 'A' in subdiff:
                 target.insert(idx, subdiff['A'])
-                offset = offset + 1
-                idx = idx + 1
+                offset += 1
+                idx += 1
             elif 'R' in subdiff:
                 if target.pop(idx) != subdiff['R']:
                     raise ValueError('Removing line does not match')
-                offset = offset - 1
+                offset -= 1
             elif 'U' in subdiff:
                 if target[idx] != subdiff['U']:
                     raise ValueError('Unchanged line does not match')
-                idx = idx + 1
+                idx += 1
             else:
                 raise ValueError('Unsupported operation')
 
@@ -615,10 +614,8 @@ class Patcher(object):
 
 
 class Iterator(object):
-    """
-    Nested diff iterator.
+    """Nested diff iterator."""
 
-    """
     def __init__(self, sort_keys=False):
         """
         Construct iterator.
@@ -668,8 +665,10 @@ class Iterator(object):
         idx = 0
 
         for item in ndiff['D']:
-            if 'I' in item:
+            try:
                 idx = item['I']
+            except KeyError:
+                pass
 
             yield ndiff, idx, item
 
@@ -703,11 +702,12 @@ class Iterator(object):
         """
         self.__iterators[type_] = method
 
-    def iterate(self, ndiff):
+    def iterate(self, ndiff, depth=0):
         """
-        Yield tuples with diff, key and subdiff for each nested diff.
+        Yield tuples with diff, key, subdiff and depth for each nested diff.
 
         :param ndiff: nested diff.
+        :param depth: depth initial value (for use in subiterators).
 
         """
         stack = [self.get_iterator(ndiff)]
@@ -717,21 +717,21 @@ class Iterator(object):
                 ndiff, key, subdiff = next(stack[-1])
             except StopIteration:
                 stack.pop()
+                depth -= 1
                 continue
 
-            yield ndiff, key, subdiff
+            yield ndiff, key, subdiff, depth
 
             if subdiff is None:
                 continue
 
+            depth += 1
             stack.append(self.get_iterator(subdiff))
 
 
 def diff(a, b, **kwargs):
     """
-    Compute recursive diff for two passed objects.
-
-    Wrapper around Differ.diff() for backward compatibility.
+    Return recursive diff for two passed objects.
 
     :param a: First object to diff.
     :param b: Second object to diff.
@@ -746,8 +746,6 @@ def patch(target, ndiff, **kwargs):
     """
     Return patched object.
 
-    Wrapper around Patcher.patch() for backward compatibility.
-
     :param target: Object to patch.
     :param ndiff: Nested diff.
 
